#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

__all__ = ["__version__", "NeptuneCallback"]

import io
import tempfile
from typing import Union

# Note: we purposefully try to import `tensorflow.keras.callbacks.Callback`
# before `keras.callbacks.Callback` because the former is compatible with both
# `tensorflow.keras` and `keras`, while the latter is only compatible
# with `keras`. See https://github.com/keras-team/keras/issues/14125

try:
    from tensorflow.keras.callbacks import Callback
    from tensorflow.keras.utils import model_to_dot
except ImportError as exc:
    try:
        from keras.callbacks import Callback
        from keras.utils import model_to_dot
    except ImportError:
        msg = """
        keras package not found.

        As Keras is now part of Tensorflow you should install it by running
            pip install tensorflow"""
        raise ModuleNotFoundError(msg) from exc

try:
    # neptune-client=0.9.0+ package structure
    import neptune.new as neptune
    from neptune.new.exceptions import NeptuneException
    from neptune.new.integrations.utils import (
        expect_not_an_experiment,
        verify_type,
    )
    from neptune.new.types import File
    from neptune.new.utils import stringify_unsupported
except ImportError:
    # neptune-client>=1.0.0 package structure
    import neptune
    from neptune.exceptions import NeptuneException
    from neptune.integrations.utils import verify_type, expect_not_an_experiment
    from neptune.types import File
    from neptune.utils import stringify_unsupported

from neptune_tensorflow_keras.impl.version import __version__

INTEGRATION_VERSION_KEY = "source_code/integrations/neptune-tensorflow-keras"


class NeptuneCallback(Callback):
    """Captures model training metadata and logs them to Neptune.

    See the example run here https://ui.neptune.ai/shared/keras-integration/e/KERAS-23/logs

    Args:
        run: Neptune run or namespace handler.
        base_namespace: Namespace (folder) under which all metadata
            logged by the NeptuneCallback will be stored. Defaults to "training".
        log_on_batch: Log the metrics also for each batch, not only each epoch.
        log_model_diagram: Save the model visualization. Defaults to False.
            This functionality requires pydot to be installed (https://pypi.org/project/pydot/).

    Example:

        Initialize Neptune client:

        >>> import neptune.new as neptune
        >>> run = neptune.init_run(
        ...     project="common/tf-keras-integration",
        ...     api_token=neptune.ANONYMOUS_API_TOKEN,
        ... )

        Instantiate the callback and pass it to the `callbacks` argument of `model.fit()`:

        >>> from neptune.new.integrations.tensorflow_keras import NeptuneCallback
        >>> neptune_callback = NeptuneCallback(run=run)
        >>> model.fit(x_train, y_train, callbacks=[neptune_callback])

    Note:
        To use this module, you need to have Keras or Tensorflow 2 installed.
    """

    def __init__(
        self,
        run: Union[neptune.Run, neptune.handler.Handler],
        base_namespace: str = "training",
        log_model_diagram: bool = False,
        log_on_batch: bool = False,
    ):
        super().__init__()

        expect_not_an_experiment(run)
        verify_type("run", run, (neptune.Run, neptune.handler.Handler))
        verify_type("base_namespace", base_namespace, (str, type(None)))
        verify_type("log_model_diagram", log_model_diagram, bool)

        self._run = run
        self._log_model_diagram = log_model_diagram

        self._log_on_batch = log_on_batch

        if base_namespace.endswith("/"):
            self._base_namespace = base_namespace[:-1]
        else:
            self._base_namespace = base_namespace

        root_obj = self._run
        if isinstance(self._run, neptune.handler.Handler):
            root_obj = self._run.get_root_object()

        root_obj[INTEGRATION_VERSION_KEY] = __version__

    @property
    def _metric_logger(self):
        return self._run[self._base_namespace]

    @property
    def _model_logger(self):
        return self._run[self._base_namespace]["model"]

    def _log_metrics(self, logs, category: str, trigger: str):
        if not logs:
            return

        logger = self._metric_logger[category][trigger]

        for metric, value in logs.items():
            try:
                if metric in ("batch", "size") or metric.startswith("val_"):
                    continue
                logger[metric].append(value)
            except NeptuneException:
                pass

    def on_train_begin(self, logs=None):
        optimizer_config = self.model.optimizer.get_config()  # it is a dict
        self._model_logger["optimizer_config"] = stringify_unsupported(optimizer_config)
        self._metric_logger["fit_params"] = self.params

    def on_train_end(self, logs=None):
        # We need this to be logged at the end of the training, otherwise we are risking this to happen:
        # https://stackoverflow.com/q/55908188/3986320
        self._model_logger["summary"] = _model_summary_file(self.model)

        if self._log_model_diagram:
            self._model_logger["visualization"] = _model_diagram(self.model)

    def on_train_batch_end(self, batch, logs=None):
        if self._log_on_batch:
            self._log_metrics(logs, "train", "batch")

    def on_epoch_begin(self, epoch, logs=None):
        self._model_logger["learning_rate"].append(self.model.optimizer.learning_rate)

    def on_epoch_end(self, epoch, logs=None):
        self._log_metrics(logs, "train", "epoch")

    def on_test_batch_end(self, batch, logs=None):
        if self._log_on_batch:
            self._log_metrics(logs, "validation", "batch")

    def on_test_end(self, logs=None):
        self._log_metrics(logs, "validation", "epoch")


def _model_summary_file(model) -> File:
    stream = io.StringIO()
    model.summary(print_fn=lambda x: stream.write(x + "\n"))
    return File.from_stream(stream, extension="txt")


def _model_diagram(model) -> File:
    dot = model_to_dot(model)
    if dot is not None:
        # the same as TF/Keras does, we will fail with ImportError unless using a notebook,
        # where it just prints a warning message
        tmp = tempfile.NamedTemporaryFile(suffix=".png", delete=False)
        dot.write(tmp.name, format="png")
        return File(tmp.name)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['neptune_tensorflow_keras', 'neptune_tensorflow_keras.impl']

package_data = \
{'': ['*']}

install_requires = \
['neptune-client>=0.16.17', 'tensorflow>2.0.0']

extras_require = \
{':python_version < "3.8"': ['importlib-metadata'],
 'dev': ['pre-commit', 'pytest>=5.0', 'pytest-cov==2.10.1', 'pydot']}

setup_kwargs = {
    'name': 'neptune-tensorflow-keras',
    'version': '2.0.0',
    'description': 'Neptune.ai tensorflow-keras integration library',
    'long_description': '# Neptune + TensorFlow/Keras Integration\n\nExperiment tracking, model registry, data versioning, and live model monitoring for TensorFlow/Keras trained models.\n\n## What will you get with this integration?\n\n* Log, display, organize, and compare ML experiments in a single place\n* Version, store, manage, and query trained models, and model building metadata\n* Record and monitor model training, evaluation, or production runs live\n* Collaborate with a team\n\n## What will be logged to Neptune?\n\n* hyperparameters for every run,\n* learning curves for losses and metrics during training,\n* hardware consumption and stdout/stderr output during training,\n* TensorFlow tensors as images to see model predictions live,\n* training code and git commit information,\n* model weights\n* [other metadata](https://docs.neptune.ai/you-should-know/what-can-you-log-and-display)\n\n![image](https://user-images.githubusercontent.com/97611089/160638338-8a276866-6ce8-4d0a-93f5-bd564d00afdf.png)\n*Example charts in the Neptune UI with logged accuracy and loss*\n\n\n## Resources\n\n* [Documentation](https://docs.neptune.ai/integrations-and-supported-tools/model-training/tensorflow-keras)\n* [Code example on GitHub](https://github.com/neptune-ai/examples/blob/main/integrations-and-supported-tools/tensorflow-keras/scripts)\n* [Runs logged in the Neptune app](https://app.neptune.ai/o/common/org/tf-keras-integration/e/TFK-18/all)\n* [Run example in Google Colab](https://colab.research.google.com/github/neptune-ai/examples/blob/master/integrations-and-supported-tools/tensorflow-keras/notebooks/Neptune_TensorFlow_Keras.ipynb)\n\n## Example\n\n```python\n# On the command line:\npip install tensorflow neptune-client neptune-tensorflow-keras\n```\n```python\n# In Python:\nimport neptune.new as neptune\nfrom neptune.new.integrations.tensorflow_keras import NeptuneCallback\n\n\n# Start a run\nrun = neptune.init(project="common/tf-keras-integration",\n                   api_token="ANONYMOUS")\n\n\n# Create a NeptuneCallback instance\nneptune_cbk = NeptuneCallback(run=run, base_namespace="metrics")\n\n\n# Pass the callback to model.fit()\nmodel.fit(x_train, y_train,\n          epochs=5,\n          batch_size=64,\n          callbacks=[neptune_cbk])\n\n\n# Stop the run\nrun.stop()\n```\n\n## Support\n\nIf you got stuck or simply want to talk to us, here are your options:\n\n* Check our [FAQ page](https://docs.neptune.ai/getting-started/getting-help#frequently-asked-questions)\n* You can submit bug reports, feature requests, or contributions directly to the repository.\n* Chat! When in the Neptune application click on the blue message icon in the bottom-right corner and send a message. A real person will talk to you ASAP (typically very ASAP),\n* You can just shoot us an email at support@neptune.ai\n',
    'author': 'neptune.ai',
    'author_email': 'contact@neptune.ai',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://neptune.ai/',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
